"""Stub (interface) file for ``_version.py``.

Provides type hint PEP 484 information for the ``_version.py``
file because it is generated by setuptools_scm and may
not exist when type checking is done and will not
contain type hint information.
"""  # noqa: PYI021

version: str
